#include "mapinc.h"

#define CHRRAM (GameMemBlock)
static uint8 latche;

DECLFW(CPROMWrite)
{
 latche=V&3;
 setvram4(0x1000,CHRRAM+((V&3)<<12));
}

static void CPROMReset(void)
{
 setprg32(0x8000,0);
 setvram8(0);
 SetReadHandler(0x8000,0xFFFF,CartBR);
 SetWriteHandler(0x8000,0xffff,CPROMWrite);
}

static void CPROMRestore(int version)
{
 setvram4(0x1000,CHRRAM+((latche)<<12));
}

void CPROM_Init(void)
{
 BoardPower=CPROMReset;
 GameStateRestore=CPROMRestore;
 AddExState(&latche, 1, 0, "LATC");
}

DECLFW(CNROMWrite)
{
 latche=V&3;
 setchr8(V&3);
}

static void CNROMReset(void)
{
  setprg16(0x8000,0);
  setprg16(0xC000,1);
  SetReadHandler(0x8000,0xFFFF,CartBR);
  SetWriteHandler(0x8000,0xffff,CNROMWrite);
}

static void CNROMRestore(int version)
{
 setchr8(latche);
}

void CNROM_Init(void)
{
 BoardPower=CNROMReset;
 GameStateRestore=CNROMRestore;
 AddExState(&latche, 1, 0, "LATC");
}

static void NROM128Reset(void)
{
  setprg16(0x8000,0);
  setprg16(0xC000,0);
  setchr8(0);
  SetReadHandler(0x8000,0xFFFF,CartBR);
}

static void NROM256Reset(void)
{
  setprg16(0x8000,0);
  setprg16(0xC000,1);
  setchr8(0);
  SetReadHandler(0x8000,0xFFFF,CartBR);
}
void NROM128_Init(void)
{
  BoardPower=NROM128Reset;
}

void NROM256_Init(void)
{
  BoardPower=NROM256Reset;
}

static DECLFW(MHROMWrite)
{
 setprg32(0x8000,V>>4);
 setchr8(V);
 latche=V;
}

static void MHROMReset(void)
{
 setprg32(0x8000,0);
 setchr8(0);
 latche=0;
 SetReadHandler(0x8000,0xFFFF,CartBR);
}

static void MHROMRestore(int version)
{
 setprg32(0x8000,latche);
 setchr8(latche);
 SetWriteHandler(0x8000,0xffff,MHROMWrite);
}

void MHROM_Init(void)
{ 
 BoardPower=MHROMReset;
 AddExState(&latche, 1, 0,"LATC");
 PRGmask32[0]&=1;
 CHRmask8[0]&=1;
 GameStateRestore=MHROMRestore;
}

static void UNROMRestore(int version)
{
 setprg16(0x8000,latche);
}

static DECLFW(UNROMWrite)
{
 setprg16(0x8000,V);
 latche=V;
}

static void UNROMReset(void)
{
 setprg16(0x8000,0);
 setprg16(0xc000,~0);
 setvram8(CHRRAM);
 SetWriteHandler(0x8000,0xffff,UNROMWrite);
 SetReadHandler(0x8000,0xFFFF,CartBR);
 latche=0;
}

void UNROM_Init(void)
{
 BoardPower=UNROMReset;
 PRGmask16[0]&=7;
 AddExState(&latche, 1, 0, "LATC");
 AddExState(CHRRAM, 8192, 0, "CHRR");
 GameStateRestore=UNROMRestore;
}
