#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sched.h>
#include <sys/soundcard.h>

#include "../../types.h"

static int format;
static int dspfd;

// fsize is in samples, not bytes(gets translated before ioctl())
int InitUNIXDSPSound(int *rate, int bits, int fsize, int nfrags)
{
 int x;

 printf("  Opening /dev/dsp...");
 dspfd=open("/dev/dsp",O_WRONLY);
 if(dspfd==-1) goto __disperror;

 if(!bits) goto skip16check;
 x=AFMT_S16_LE;
 format=0;
 printf("\n   Setting format to 16-bit, signed, LSB first...");
 if(ioctl(dspfd,SNDCTL_DSP_SETFMT,&x)==-1)
 {
  skip16check:
  x=AFMT_U8;
  printf("\n   Setting format to 8-bit, unsigned...");
  if(ioctl(dspfd,SNDCTL_DSP_SETFMT,&x)==-1) goto __disperror;
  format=1;
 }

 printf("\n   Setting fragment size to %d samples and number of fragments to %d...",1<<fsize,nfrags);

 if(!format)
  fsize++;
 x=fsize|(nfrags<<16);

 if(ioctl(dspfd,SNDCTL_DSP_SETFRAGMENT,&x)==-1)
  printf("ERROR (continuing anyway)\n");
 x=1;
 printf("\n   Setting stereo sound...");  
 if(ioctl(dspfd,SNDCTL_DSP_STEREO,&x)==-1) goto __disperror;
 printf("\n   Setting playback rate of %d hz...",*rate);
 if(ioctl(dspfd,SNDCTL_DSP_SPEED,rate)==-1) goto __disperror;
 printf("Set to %d hz\n",*rate);
 if(*rate<8192 || *rate>65535)
 {
  printf("    Sample rate is out of the acceptable range(8192-65535).\n");
  close(dspfd);
  return(0);
 }
 return 1;
 __disperror:
 printf("ERROR\n");
 return 0;
}

void KillUNIXDSPSound(void)
{
  close(dspfd);
}

static uint32 WBuffer[2048];
void WriteUNIXDSPSound(int32 *Buffer, int Count, int noblocking)
{
 int P,c;
 int32 *src=Buffer;

 if(format)
 {
  uint8 *dest=(uint16 *)WBuffer;
  for(P=Count;P;P--,dest++,src++)
   *dest= ( ((uint8)((*src)>>8)^128) << 8 ) | ( (uint8)((*src)>>8)^128 );
  c=Count;
 }
 else
 {
  uint32 *dest=WBuffer;
  for(P=Count;P;P--,dest++,src++)
   *dest=( ((uint16)(int16)(*src)) << 16 ) | ((uint16)(int16)(*src));
  c=Count<<1;
 }

 //noblocking=!noblocking; // speed testing
 if(noblocking)
 {
  struct audio_buf_info ai;
  if(!ioctl(dspfd,SNDCTL_DSP_GETOSPACE,&ai))
   if(ai.bytes<c)
    return;
 }
 write(dspfd,(uint16 *)WBuffer,c<<1);
}
