#include "qpe_fceu.h"

#include <qpe/qpeapplication.h>
#include <qpe/qpemenubar.h>
#include <qmenubar.h>
#include <qpopupmenu.h>
#include <qaction.h>
//#include <qapplication.h>
#include "fileBrowser.h"
#include <qmessagebox.h>

QpeFceu* qpe_fceu;
QPEApplication* app;

int main( int argc, char **argv )
{
    QPEApplication a( argc, argv );

    QPEApplication::grabKeyboard();
         
    QpeFceu e;
    QPEApplication::setInputMethodHint( &e, QPEApplication::AlwaysOff );
    qpe_fceu=&e;
    app=&a;
    a.showMainDocumentWidget(&e);
    e.show();

    e.init();
    a.exec();
}

uchar* fb;

QpeFceu::QpeFceu( QWidget *parent = 0, const char *name = 0, WFlags f = 0 ) : QMainWindow(parent, name, f)
{
	screen=new QImage(256,240,16);
	printf("screen 0x%x\n",(unsigned int)screen);
	painter=new QDirectPainter(this);
	fb=painter->frameBuffer();
	delete painter;
	setFocusPolicy(StrongFocus);

        QMenuBar *mb = new QMenuBar( this );
        mb->setMargin(0);
	mb->setFixedHeight(18);
        QPopupMenu *file = new QPopupMenu( this );
        connect(file,SIGNAL(aboutToShow()),this,SLOT(disabledraw()));
        connect(file,SIGNAL(aboutToHide()),this,SLOT(enabledraw()));

	QAction *a = new QAction( tr( "Open" ), QString::null, 0, this, 0 );
	connect( a, SIGNAL( activated() ), this, SLOT( openselector() ) );
	a->addTo( file );
        
        a = new QAction( tr( "Reset" ), QString::null, 0, this, 0 );
	connect( a, SIGNAL( activated() ), this, SLOT( reset() ) );
	a->addTo( file );

	a = new QAction( tr( "Quit" ), QString::null, 0, this, 0 );
	connect( a, SIGNAL( activated() ), this, SLOT( quit() ) );
	a->addTo( file );

	mb->insertItem( tr( "File" ), file );


        QPopupMenu *options = new QPopupMenu( this );
        connect(options,SIGNAL(aboutToShow()),this,SLOT(disabledraw()));
        connect(options,SIGNAL(aboutToHide()),this,SLOT(enabledraw()));
        a = new QAction( tr( "Sound" ), QString::null, 0, this, 0, TRUE );
        a->setOn(TRUE);
	connect( a, SIGNAL( toggled(bool) ), this, SLOT( sound_tgl(bool) ) );
	a->addTo( options );

	a = new QAction( tr( "Sample rate" ), QString::null, 0, this, 0);
	connect( a, SIGNAL( activated() ), this, SLOT( change_rate() ) );
	a->addTo( options );

	a = new QAction( tr( "Frameskip +" ), QString::null, 0, this, 0);
	connect( a, SIGNAL( activated() ), this, SLOT( frameskip_inc() ) );
	a->addTo( options );

	a = new QAction( tr( "Frameskip -" ), QString::null, 0, this, 0);
	connect( a, SIGNAL( activated() ), this, SLOT( frameskip_dec() ) );
	a->addTo( options );

	a = new QAction( tr( "Rotate" ), QString::null, 0, this, 0, TRUE);
	connect( a, SIGNAL( toggled(bool) ), this, SLOT( rotate(bool) ) );
	a->addTo( options );

	mb->insertItem( tr( "Options" ), options );

	setCaption("FCE Ultra");

	frameskip_lbl=new QLabel(this);
	frameskip_lbl->setGeometry(0,18,236,10);
	frameskip_lbl->setAlignment(AlignRight);
	QFont f=frameskip_lbl->font();
	f.setPointSize(4);
	frameskip_lbl->setFont(f);
}

void QpeFceu::init()	
{
	sound=true;
	enabledraw();
	frameskip=3;
	rotated=false;

	JS=0;
	
	FCEUI_Initialize();
	DriverInputInterface(DII_SETCONTROL,0,1,&JS);
        rate=11025;
         if(InitUNIXDSPSound(&rate,16,8,8))
          FCEUI_Sound(rate);
         else
          sound=false;
	updateStatusLine();
}

QpeFceu::~QpeFceu()
{
	// Qt should handle
	FCEUI_CloseGame();
	 if(sound)
	   KillUNIXDSPSound();
	   
	delete screen;
}

void QpeFceu::reset()
{
	DriverInterface(DES_RESET,0);
}

void QpeFceu::quit()
{
	FCEUI_CloseGame();
	exit(0);
}

void QpeFceu::openselector()
{
	FCEUI_CloseGame();
	disabledraw();
	fileBrowser *browseForFiles;
	browseForFiles=new fileBrowser(this,"fileBrowser",TRUE,0, "*.nes;*.fds;*.nsf;*.unf");
	browseForFiles->exec();
	QString file=browseForFiles->selectedFileName;
	delete browseForFiles;
	if(file.isEmpty()) return;
	if(FCEUI_LoadGame((char*)file.latin1()))
	{
		reset();
		enabledraw();
		FCEUI_Emulate();
	}
}

void QpeFceu::sound_tgl(bool toggle)
{
	if(toggle)
	{
		if(sound==false)
		{
		         if(InitUNIXDSPSound(&rate,16,8,8)) {
			  sound=true;
		          FCEUI_Sound(rate);
			}
		         else
		          sound=false;
		}
	}
	else
	{
		if(sound==true)
		{
			KillUNIXDSPSound();
			sound=false;
		}
	}
	updateStatusLine();
}

void QpeFceu::frameskip_inc()
{
	++frameskip;
	updateStatusLine();
}

void QpeFceu::frameskip_dec()
{
	if(frameskip>0)
	{
		--frameskip;
		updateStatusLine();
	}
}

void QpeFceu::change_rate()
{
	if(rate < 15000)
		rate=22050;
	else
		rate=11025;
	if(sound==true)
	{
		KillUNIXDSPSound();
		if(InitUNIXDSPSound(&rate,16,8,8)) {
			sound=true;
			FCEUI_Sound(rate);
		}
		else
			sound=false;
	}
	updateStatusLine();
}

void QpeFceu::updateStatusLine()
{
	frameskip_lbl->setText(
		QString("Sound ") +
		(sound?QString("on"):QString("off")) +
		QString(" at ")+QString::number(rate)+
		QString(" Hz     Frameskip ")+
		QString::number(frameskip)
	);
}

void QpeFceu::rotate(bool r)
{
	rotated=r;
}

void QpeFceu::keyPressEvent(QKeyEvent* event)
{
	//printf("pressevent %d\n",event->key());
	// handle keypresses
	switch(event->key())
	{
		case Key_Up:
			JS|=(rotated?JOY_RIGHT:JOY_UP);
			break;
		case Key_Down:
			JS|=(rotated?JOY_LEFT:JOY_DOWN);
			break;
		case Key_Left:
			JS|=(rotated?JOY_UP:JOY_LEFT);
			break;
		case Key_Right:
			JS|=(rotated?JOY_DOWN:JOY_RIGHT);
			break;
		case Key_F9:
			JS|=JOY_SELECT;
			break;
		case Key_F10:
			JS|=JOY_START;
			break;
		case Key_F11:
			JS|=JOY_B;
			break;
		case Key_F12:
			JS|=JOY_A;
			break;
		//default:
		//	event->ignore();
		//	return;
	}
	event->accept();
}

void QpeFceu::keyReleaseEvent(QKeyEvent* event)
{
	//printf("releaseevent %d\n",event->key());
	// handle releases
	switch(event->key())
	{
		case Key_Up:
			JS&=~(rotated?JOY_RIGHT:JOY_UP);
			break;
		case Key_Down:
			JS&=~(rotated?JOY_LEFT:JOY_DOWN);
			break;
		case Key_Left:
			JS&=~(rotated?JOY_UP:JOY_LEFT);
			break;
		case Key_Right:
			JS&=~(rotated?JOY_DOWN:JOY_RIGHT);
			break;
		case Key_F9:
			JS&=~JOY_SELECT;
			break;
		case Key_F10:
			JS&=~JOY_START;
			break;
		case Key_F11:
			JS&=~JOY_B;
			break;
		case Key_F12:
			JS&=~JOY_A;
			break;
		//default:
		//	event->ignore();
		//	return;
	}
	event->accept();
}

void QpeFceu::closeEvent(QCloseEvent* event)
{
	event->accept();
	disabledraw();
	FCEUI_CloseGame();
}

void QpeFceu::enabledraw()
{
	drawenabled=true;
}

void QpeFceu::disabledraw()
{
	drawenabled=false;
}

extern "C" {

void FCEUD_PrintError(char* s)
{
	QMessageBox::warning(0,"FCE Ultra error",s);
}

void FCEUD_SendMessage(int msg)
{
    switch (msg) {
    case DSM_RESETSPECIAL:
        qpe_fceu->gametype = 0;
        break;
    case DSM_SPECIALSYSTEM_VSUNISYSTEM:
    case DSM_SPECIALSYSTEM_NSF:
    case DSM_SPECIALSYSTEM_FDS:
        qpe_fceu->gametype = msg;
        break;
    }
}

void FCEUD_UpdateInput(void)
{
	app->processEvents();
}

void FCEUD_WriteSoundData(int32 *Buffer, int Count)
{
	WriteUNIXDSPSound(Buffer,Count,0);
}

static uint8 palette[256][3];
static uint16 palette565[256];

void FCEUD_SetPalette(uint8 index, uint8 r, uint8 g, uint8 b)
{
	palette[index][0]=r;
	palette[index][1]=g;
	palette[index][2]=b;
	palette565[index]=(b>>3) + ((g<<3)&0x7e0) + ((r<<8)&0xf800);
}

void FCEUD_GetPalette(uint8 index, uint8 *r, uint8 *g, uint8 *b)
{
	*r=palette[index][0];
	*g=palette[index][1];
	*b=palette[index][2];
}

void FCEUD_BlitScreen(uint8 *XBuf)
{
	FCEUI_FrameSkip(qpe_fceu->frameskip);
	if(!qpe_fceu->drawenabled) return;
	
	if(!qpe_fceu->rotated)
	{
		for(int j=8;j<248;j++)
		{
			for(int i=239; i; i--)
			{
				*((unsigned short*)fb+(256+20)-i+(j-8)*320)=palette565[*(XBuf+i*272+j)];  //scol;
			}
		}
	}
	else
	{
		for(int i=1;i<240;i++)
		{
			for(int j=0;j<256;j++)
			{
				*((unsigned short*)fb+j+20+i*320)=palette565[*(XBuf+i*272+j)];
			}
		}
	}
}

} // extern "C"
